@@ -391,6 +391,11 @@ impl Shape {
         self.0
     }
 
+    /// Return the origin [`Coord`] of this `Shape`.
+    pub fn origin(&self) -> Coord {
+        iter::repeat(0).take(self.len()).collect()
+    }
+
     /// Return the number of elements contained within this `Shape`.
     pub fn size(&self) -> u64 {
         self.0.iter().product()
@@ -452,6 +457,13 @@ impl From<Vec<u64>> for Shape {
     }
 }
 
+impl FromIterator<u64> for Shape {
+    fn from_iter<I: IntoIterator<Item = u64>>(iter: I) -> Self {
+        let dims = Vec::<u64>::from_iter(iter);
+        Self(dims)
+    }
+}
+
 impl TryCastFrom<Value> for Shape {
     fn can_cast_from(value: &Value) -> bool {
         value.matches::<Vec<u64>>()
