@@ -1,3 +1,4 @@
+use std::convert::TryFrom;
 use std::fmt;
 use std::iter::{self, FromIterator};
 use std::marker::PhantomData;
@@ -9,35 +10,55 @@ use destream::de;
 use futures::stream::{self, Stream, StreamExt, TryStreamExt};
 use futures::{future, try_join, TryFutureExt};
 use log::debug;
-use number_general::{Number, NumberClass, NumberInstance, NumberType};
 use strided::Stride;
 
+use tc_btree::Node;
 use tc_error::*;
 use tc_transact::fs::{BlockData, BlockId, CopyFrom, Dir, File, Persist, Restore};
 use tc_transact::{Transact, Transaction, TxnId};
-use tc_value::Value;
-use tcgeneric::{TCBoxTryFuture, TCTryStream};
+use tc_value::{Number, NumberClass, NumberInstance, NumberType, Value};
+use tcgeneric::{TCBoxTryFuture, TCTryStream, Tuple};
 
+use crate::stream::{block_offsets, coord_block, coord_bounds, Read, ReadValueAt};
 use crate::transform::{self, Rebase};
-use crate::{Bounds, Coord, Read, ReadValueAt, Schema, Shape, TensorAccess};
+use crate::{Bounds, Coord, Schema, Shape, TensorAccess, TensorType};
 
 use super::access::BlockListTranspose;
-use super::{block_offsets, coord_block, DenseAccess, DenseAccessor, PER_BLOCK};
+use super::{DenseAccess, DenseAccessor, PER_BLOCK};
 
 const MEBIBYTE: usize = 1_048_576;
 
 /// A wrapper around a `DenseTensor` [`File`]
 #[derive(Clone)]
-pub struct BlockListFile<F, D, T> {
-    file: F,
+pub struct BlockListFile<FD, FS, D, T> {
+    file: FD,
     schema: Schema,
+    sparse: PhantomData<FS>,
     dir: PhantomData<D>,
     txn: PhantomData<T>,
 }
 
-impl<F: File<Array>, D: Dir, T: Transaction<D>> BlockListFile<F, D, T> {
+impl<FD, FS, D, T> BlockListFile<FD, FS, D, T> {
+    fn new(file: FD, schema: Schema) -> Self {
+        Self {
+            file,
+            schema,
+            sparse: PhantomData,
+            dir: PhantomData,
+            txn: PhantomData,
+        }
+    }
+}
+
+impl<FD, FS, D, T> BlockListFile<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+{
     /// Construct a new `BlockListFile` with the given [`Shape`], filled with the given `value`.
-    pub async fn constant(file: F, txn_id: TxnId, shape: Shape, value: Number) -> TCResult<Self> {
+    pub async fn constant(file: FD, txn_id: TxnId, shape: Shape, value: Number) -> TCResult<Self> {
         if !file.is_empty(&txn_id).await? {
             return Err(TCError::unsupported(
                 "cannot create new tensor: file is not empty",
@@ -53,41 +74,72 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> BlockListFile<F, D, T> {
         let trailing_len = (size % PER_BLOCK as u64) as usize;
         if trailing_len > 0 {
             let blocks = blocks.chain(iter::once(Ok(Array::constant(value.clone(), trailing_len))));
-            BlockListFile::from_blocks(file, txn_id, shape, value.class(), stream::iter(blocks))
-                .await
+            Self::from_blocks(
+                file,
+                txn_id,
+                Some(shape),
+                value.class(),
+                stream::iter(blocks),
+            )
+            .await
         } else {
-            BlockListFile::from_blocks(file, txn_id, shape, value.class(), stream::iter(blocks))
-                .await
+            Self::from_blocks(
+                file,
+                txn_id,
+                Some(shape),
+                value.class(),
+                stream::iter(blocks),
+            )
+            .await
         }
     }
 
     /// Construct a new `BlockListFile` from the given `Stream` of [`Array`] blocks.
     pub async fn from_blocks<S: Stream<Item = TCResult<Array>> + Send + Unpin>(
-        file: F,
+        file: FD,
         txn_id: TxnId,
-        shape: Shape,
+        shape: Option<Shape>,
         dtype: NumberType,
         blocks: S,
     ) -> TCResult<Self> {
-        blocks
+        let size = blocks
             .enumerate()
             .map(|(i, r)| r.map(|block| (BlockId::from(i), block)))
-            .map_ok(|(id, block)| file.create_block(txn_id, id, block))
+            .map_ok(|(id, block)| {
+                let len = block.len() as u64;
+                file.create_block(txn_id, id, block).map_ok(move |_| len)
+            })
             .try_buffer_unordered(num_cpus::get())
-            .try_fold((), |_, _| future::ready(Ok(())))
+            .try_fold(0u64, |block_len, size| future::ready(Ok(size + block_len)))
             .await?;
 
-        Ok(BlockListFile {
-            file,
-            schema: (shape, dtype),
-            dir: PhantomData,
-            txn: PhantomData,
-        })
+        if file.is_empty(&txn_id).await? {
+            return Err(TCError::unsupported(
+                "tried to create a dense tensor from an empty block list",
+            ));
+        }
+
+        let shape = if let Some(shape) = shape {
+            if shape.size() != size {
+                return Err(TCError::unsupported(format!(
+                    "dense tensor of shape {} requires {} elements but found {}",
+                    shape,
+                    shape.size(),
+                    size
+                )));
+            }
+
+            shape
+        } else {
+            vec![size].into()
+        };
+
+        Ok(Self::new(file, Schema { shape, dtype }))
     }
 
     /// Construct a new `BlockListFile` from the given `Stream` of elements.
     pub async fn from_values<S: Stream<Item = Number> + Send + Unpin>(
-        file: F,
+        file: FD,
         txn_id: TxnId,
         shape: Shape,
         dtype: NumberType,
@@ -102,17 +154,12 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> BlockListFile<F, D, T> {
             i += 1;
         }
 
-        Ok(BlockListFile {
-            file,
-            schema: (shape, dtype),
-            dir: PhantomData,
-            txn: PhantomData,
-        })
+        Ok(Self::new(file, Schema { shape, dtype }))
     }
 
     /// Construct a new `BlockListFile` of elements evenly distributed between `start` and `stop`.
     pub async fn range(
-        file: F,
+        file: FD,
         txn_id: TxnId,
         shape: Shape,
         start: Number,
@@ -146,7 +193,7 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> BlockListFile<F, D, T> {
     /// Sort the elements in this `BlockListFile`.
     pub async fn merge_sort(&self, txn_id: TxnId) -> TCResult<()> {
         let num_blocks = div_ceil(self.size(), PER_BLOCK as u64);
-        if num_blocks == 1 {
+        if num_blocks <= 1 {
             let block_id = BlockId::from(0u64);
             let mut block = self.file.write_block(txn_id, block_id).await?;
             block.sort(true)?;
@@ -173,30 +220,37 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> BlockListFile<F, D, T> {
     }
 }
 
-impl<F: Send, D: Send, T: Send> TensorAccess for BlockListFile<F, D, T> {
+impl<FD: Send, FS: Send, D: Send, T: Send> TensorAccess for BlockListFile<FD, FS, D, T> {
     fn dtype(&self) -> NumberType {
-        self.schema.1
+        self.schema.dtype
     }
 
     fn ndim(&self) -> usize {
-        self.schema.0.len()
+        self.schema.shape.len()
     }
 
     fn shape(&'_ self) -> &'_ Shape {
-        &self.schema.0
+        &self.schema.shape
     }
 
     fn size(&self) -> u64 {
-        self.schema.0.size()
+        self.schema.shape.size()
     }
 }
 
 #[async_trait]
-impl<F: File<Array>, D: Dir, T: Transaction<D>> DenseAccess<F, D, T> for BlockListFile<F, D, T> {
-    type Slice = BlockListFileSlice<F, D, T>;
-    type Transpose = BlockListTranspose<F, D, T, Self>;
+impl<FD, FS, D, T> DenseAccess<FD, FS, D, T> for BlockListFile<FD, FS, D, T>
+where
+    FD: File<Array> + TryFrom<D::File, Error = TCError>,
+    FS: File<Node> + TryFrom<D::File, Error = TCError>,
+    D: Dir,
+    T: Transaction<D>,
+    D::FileClass: From<TensorType>,
+{
+    type Slice = BlockListFileSlice<FD, FS, D, T>;
+    type Transpose = BlockListTranspose<FD, FS, D, T, Self>;
 
-    fn accessor(self) -> DenseAccessor<F, D, T> {
+    fn accessor(self) -> DenseAccessor<FD, FS, D, T> {
         DenseAccessor::File(self)
     }
 
@@ -207,7 +261,9 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> DenseAccess<F, D, T> for BlockLi
             let block_stream = Box::pin(
                 stream::iter(0..(div_ceil(size, PER_BLOCK as u64)))
                     .map(BlockId::from)
-                    .then(move |block_id| file.clone().read_block_owned(*txn.id(), block_id))
+                    .then(move |block_id| {
+                        file.clone().read_block_owned(*txn.id(), block_id)
+                    })
                     .map_ok(|block| (*block).clone()),
             );
 
@@ -280,12 +336,10 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> DenseAccess<F, D, T> for BlockLi
 
     fn write_value_at(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCBoxTryFuture<()> {
         Box::pin(async move {
-            debug!("BlockListFile::write_value_at {:?} <- {}", coord, value);
-
             if !self.shape().contains_coord(&coord) {
                 return Err(TCError::bad_request(
-                    "Invalid coordinate",
-                    format!("[{:?}]", coord),
+                    "invalid coordinate",
+                    Tuple::from(coord),
                 ));
             }
 
@@ -301,7 +355,6 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> DenseAccess<F, D, T> for BlockLi
             let mut block = self.file.write_block(txn_id, block_id).await?;
 
             let offset = offset % PER_BLOCK as u64;
-            debug!("offset is {}", offset);
 
             (*block)
                 .set_value(offset as usize, value)
@@ -310,17 +363,18 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> DenseAccess<F, D, T> for BlockLi
     }
 }
 
-impl<F: File<Array>, D: Dir, T: Transaction<D>> ReadValueAt<D> for BlockListFile<F, D, T> {
+impl<FD, FS, D, T> ReadValueAt<D> for BlockListFile<FD, FS, D, T>
+where
+    FD: File<Array> + TryFrom<D::File, Error = TCError>,
+    FS: File<Node> + TryFrom<D::File, Error = TCError>,
+    D: Dir,
+    T: Transaction<D>,
+    D::FileClass: From<TensorType>,
+{
     type Txn = T;
 
     fn read_value_at<'a>(self, txn: T, coord: Coord) -> Read<'a> {
         Box::pin(async move {
-            debug!(
-                "read value at {:?} from BlockListFile with shape {}",
-                coord,
-                self.shape()
-            );
-
             if !self.shape().contains_coord(&coord) {
                 return Err(TCError::bad_request(
                     "Coordinate is out of bounds",
@@ -334,17 +388,8 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> ReadValueAt<D> for BlockListFile
                 .map(|(d, x)| d * x)
                 .sum();
 
-            debug!("coord {:?} is offset {}", coord, offset);
-
             let block_id = BlockId::from(offset / PER_BLOCK as u64);
             let block = self.file.read_block(*txn.id(), block_id).await?;
-
-            debug!(
-                "read offset {} from block of length {}",
-                offset % PER_BLOCK as u64,
-                block.len()
-            );
-
             let value = block.get_value((offset % PER_BLOCK as u64) as usize);
 
             Ok((coord, value))
@@ -353,7 +398,13 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> ReadValueAt<D> for BlockListFile
 }
 
 #[async_trait]
-impl<F: File<Array> + Transact, D: Dir, T: Transaction<D>> Transact for BlockListFile<F, D, T> {
+impl<FD, FS, D, T> Transact for BlockListFile<FD, FS, D, T>
+where
+    FD: File<Array> + Transact,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+{
     async fn commit(&self, txn_id: &TxnId) {
         self.file.commit(txn_id).await
     }
@@ -364,22 +415,33 @@ impl<F: File<Array> + Transact, D: Dir, T: Transaction<D>> Transact for BlockLis
 }
 
 #[async_trait]
-impl<F: File<Array>, D: Dir, T: Transaction<D>, B: DenseAccess<F, D, T>> CopyFrom<D, B>
-    for BlockListFile<F, D, T>
+impl<FD, FS, D, T, B> CopyFrom<D, B> for BlockListFile<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    B: DenseAccess<FD, FS, D, T>,
 {
-    async fn copy_from(other: B, file: F, txn: T) -> TCResult<Self> {
+    async fn copy_from(other: B, file: FD, txn: T) -> TCResult<Self> {
         let txn_id = *txn.id();
         let dtype = other.dtype();
         let shape = other.shape().clone();
         let blocks = other.block_stream(txn).await?;
-        Self::from_blocks(file, txn_id, shape, dtype, blocks).await
+        Self::from_blocks(file, txn_id, Some(shape), dtype, blocks).await
     }
 }
 
 #[async_trait]
-impl<F: File<Array>, D: Dir, T: Transaction<D>> Persist<D> for BlockListFile<F, D, T> {
+impl<FD, FS, D, T> Persist<D> for BlockListFile<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+{
     type Schema = Schema;
-    type Store = F;
+    type Store = FD;
     type Txn = T;
 
     fn schema(&self) -> &Schema {
@@ -387,29 +449,30 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> Persist<D> for BlockListFile<F,
     }
 
     async fn load(_txn: &T, schema: Self::Schema, file: Self::Store) -> TCResult<Self> {
-        Ok(Self {
-            file,
-            schema,
-            dir: PhantomData,
-            txn: PhantomData,
-        })
+        Ok(Self::new(file, schema))
     }
 }
 
 #[async_trait]
-impl<F: File<Array>, D: Dir, T: Transaction<D>> Restore<D> for BlockListFile<F, D, T> {
+impl<FD, FS, D, T> Restore<D> for BlockListFile<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+{
     async fn restore(&self, backup: &Self, txn_id: TxnId) -> TCResult<()> {
-        if self.schema.0 != backup.schema.0 {
+        if self.schema.shape != backup.schema.shape {
             return Err(TCError::bad_request(
                 "cannot restore a dense Tensor from a backup with a different shape",
-                &backup.schema.0,
+                &backup.schema.shape,
             ));
         }
 
-        if self.schema.1 != backup.schema.1 {
+        if self.schema.dtype != backup.schema.dtype {
             return Err(TCError::bad_request(
                 "cannot restore a dense Tensor from a backup with a different data type",
-                &backup.schema.1,
+                &backup.schema.dtype,
             ));
         }
 
@@ -419,17 +482,23 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> Restore<D> for BlockListFile<F,
 }
 
 #[async_trait]
-impl<F: File<Array>, D: Send, T: Send> de::FromStream for BlockListFile<F, D, T> {
-    type Context = (TxnId, F, Schema);
+impl<FD, FS, D, T> de::FromStream for BlockListFile<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: Send,
+    D: Send,
+    T: Send,
+{
+    type Context = (TxnId, FD, Schema);
 
     async fn from_stream<De: de::Decoder>(
-        cxt: (TxnId, F, Schema),
+        cxt: (TxnId, FD, Schema),
         decoder: &mut De,
     ) -> Result<Self, De::Error> {
-        let (txn_id, file, (shape, dtype)) = cxt;
+        let (txn_id, file, schema) = cxt;
         let visitor = BlockListVisitor::new(txn_id, &file);
 
-        use number_general::{
+        use tc_value::{
             ComplexType as CT, FloatType as FT, IntType as IT, NumberType as NT, UIntType as UT,
         };
 
@@ -440,7 +509,7 @@ impl<F: File<Array>, D: Send, T: Send> de::FromStream for BlockListFile<F, D, T>
             ))
         }
 
-        let size = match dtype {
+        let size = match schema.dtype {
             NT::Bool => decoder.decode_array_bool(visitor).await,
             NT::Complex(ct) => match ct {
                 CT::C32 => {
@@ -477,17 +546,12 @@ impl<F: File<Array>, D: Send, T: Send> de::FromStream for BlockListFile<F, D, T>
             NT::Number => Err(err_nonspecific(NT::Number)),
         }?;
 
-        if size == shape.size() {
-            Ok(Self {
-                file,
-                schema: (shape, dtype),
-                dir: PhantomData,
-                txn: PhantomData,
-            })
+        if size == schema.shape.size() {
+            Ok(Self::new(file, schema))
         } else {
             Err(de::Error::custom(format!(
                 "tensor data has the wrong number of elements ({}) for shape {}",
-                size, shape
+                size, &schema.shape
             )))
         }
     }
@@ -709,19 +773,31 @@ impl<'a, F: File<Array>> de::Visitor for ComplexBlockListVisitor<'a, F> {
 
 /// A slice of a [`BlockListFile`]
 #[derive(Clone)]
-pub struct BlockListFileSlice<F, D, T> {
-    source: BlockListFile<F, D, T>,
+pub struct BlockListFileSlice<FD, FS, D, T> {
+    source: BlockListFile<FD, FS, D, T>,
     rebase: transform::Slice,
 }
 
-impl<F: File<Array>, D: Dir, T: Transaction<D>> BlockListFileSlice<F, D, T> {
-    fn new(source: BlockListFile<F, D, T>, bounds: Bounds) -> TCResult<Self> {
+impl<FD, FS, D, T> BlockListFileSlice<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+{
+    fn new(source: BlockListFile<FD, FS, D, T>, bounds: Bounds) -> TCResult<Self> {
         let rebase = transform::Slice::new(source.shape().clone(), bounds)?;
         Ok(Self { source, rebase })
     }
 }
 
-impl<F: File<Array>, D: Dir, T: Transaction<D>> TensorAccess for BlockListFileSlice<F, D, T> {
+impl<FD, FS, D, T> TensorAccess for BlockListFileSlice<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+{
     fn dtype(&self) -> NumberType {
         self.source.dtype()
     }
@@ -740,26 +816,30 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> TensorAccess for BlockListFileSl
 }
 
 #[async_trait]
-impl<F: File<Array>, D: Dir, T: Transaction<D>> DenseAccess<F, D, T>
-    for BlockListFileSlice<F, D, T>
+impl<FD, FS, D, T> DenseAccess<FD, FS, D, T> for BlockListFileSlice<FD, FS, D, T>
+where
+    FD: File<Array> + TryFrom<D::File, Error = TCError>,
+    FS: File<Node> + TryFrom<D::File, Error = TCError>,
+    D: Dir,
+    T: Transaction<D>,
+    D::FileClass: From<TensorType>,
 {
     type Slice = Self;
-    type Transpose = BlockListTranspose<F, D, T, Self>;
+    type Transpose = BlockListTranspose<FD, FS, D, T, Self>;
 
-    fn accessor(self) -> DenseAccessor<F, D, T> {
+    fn accessor(self) -> DenseAccessor<FD, FS, D, T> {
         DenseAccessor::Slice(self)
     }
 
-    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Number>> {
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
         let txn_id = *txn.id();
         let file = self.source.file;
-        let shape = self.source.schema.0;
+        let shape = self.source.schema.shape;
         let mut bounds = self.rebase.bounds().clone();
         bounds.normalize(&shape);
         let coord_bounds = coord_bounds(&shape);
 
         let values = stream::iter(bounds.affected())
-            .inspect(|coord| debug!("reading value from source coord {:?}", coord))
             .chunks(PER_BLOCK)
             .then(move |coords| {
                 let ndim = coords[0].len();
@@ -777,33 +857,20 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> DenseAccess<F, D, T>
                     let mut start = 0.0f64;
                     let mut values = vec![];
                     for block_id in block_ids {
-                        debug!("block {} starts at {}", block_id, start);
-
                         let (block_offsets, new_start) =
                             block_offsets(&af_indices, &af_offsets, start, block_id);
 
-                        debug!("reading {} block_offsets", block_offsets.elements());
-                        match file_clone
-                            .clone()
-                            .read_block_owned(txn_id, block_id.into())
-                            .await
-                        {
-                            Ok(block) => {
-                                values.extend(block.get(&block_offsets.into()).to_vec());
-                            }
-                            Err(cause) => return stream::iter(vec![Err(cause)]),
-                        }
-
+                        let block = file_clone.read_block(txn_id, block_id.into()).await?;
+                        values.extend(block.get(&block_offsets.into()).to_vec());
                         start = new_start;
                     }
 
-                    let values: Vec<TCResult<Number>> = values.into_iter().map(Ok).collect();
-                    stream::iter(values)
+                    Ok(Array::from(values))
                 })
             });
 
-        let values: TCTryStream<Number> = Box::pin(values.flatten());
-        Box::pin(future::ready(Ok(values)))
+        let blocks: TCTryStream<Array> = Box::pin(values);
+        Box::pin(future::ready(Ok(blocks)))
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
@@ -836,7 +903,14 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> DenseAccess<F, D, T>
     }
 }
 
-impl<F: File<Array>, D: Dir, T: Transaction<D>> ReadValueAt<D> for BlockListFileSlice<F, D, T> {
+impl<FD, FS, D, T> ReadValueAt<D> for BlockListFileSlice<FD, FS, D, T>
+where
+    FD: File<Array> + TryFrom<D::File, Error = TCError>,
+    FS: File<Node> + TryFrom<D::File, Error = TCError>,
+    D: Dir,
+    T: Transaction<D>,
+    D::FileClass: From<TensorType>,
+{
     type Txn = T;
 
     fn read_value_at<'a>(self, txn: Self::Txn, coord: Coord) -> Read<'a> {
@@ -848,13 +922,6 @@ impl<F: File<Array>, D: Dir, T: Transaction<D>> ReadValueAt<D> for BlockListFile
     }
 }
 
-#[inline]
-fn coord_bounds(shape: &Shape) -> Coord {
-    (0..shape.len())
-        .map(|axis| shape[axis + 1..].iter().product())
-        .collect()
-}
-
 #[inline]
 fn div_ceil(l: u64, r: u64) -> u64 {
     if l % r == 0 {
